# Copyright 2022 Zeeland(https://github.com/Undertone0809/). All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent.futures import ThreadPoolExecutor
from typing import Optional, List, Callable

__all__ = ['broadcast_service', 'BroadcastService']


class BroadcastService:
    """
    This class implements broadcast mode, you can import the instance by single class.
    By BroadcastService, you can send topic message,it will automatically execute the
    callback function if some classes subscribe the topic.

    example:
    ---------------------------------------------------------------------------------
    from broadcast_service import broadcast_service

    def handle_msg(params):
        print(params)

    if __name__ == '__main__':
        info = 'This is very important msg'

        # listen topic
        broadcast_service.subscribe('Test', handle_msg)

        # publish broadcast
        broadcast_service.publish('Test', info)
    ---------------------------------------------------------------------------------
    """

    def __init__(self):
        """
        pubsub_channels is the dict to store publish/subscribe data.
        pubsub_channels example:

        pubsub_channels = {
            'my_topic': [callback_function1: Callable,callback_function2: Callable]
            '__all__': [callback_function3: Callable]
        }
        """
        self.pubsub_channels: dict = {
            '__all__': []
        }
        self.enable_async: bool = True
        self.thread_pool = ThreadPoolExecutor(max_workers=5)

        # function renaming
        self.subscribe = self.listen
        self.on = self.listen
        self.publish = self.broadcast
        self.emit = self.broadcast
        self.unsubscribe = self.stop_listen
        self.off = self.stop_listen
        self.on_subscribe = self.on_listen
        self.subscribe_all = self.listen_all
        self.publish_all = self.broadcast_all

    def listen(self, topics: str or List[str], callback: Callable):
        """
        listen topic.
        """
        if type(topics) == str:
            self._listen_topic(topics, callback)
        elif type(topics) == list:
            for topic in topics:
                self._listen_topic(topic, callback)
        else:
            raise ValueError("Unknown broadcast-service error, please submit "
                             "issue to https://github.com/Undertone0809/broadcast-service/issues")

    def listen_all(self, callback: Callable):
        """
        '__all__' is a special topic. It can receive any topic message.
        """
        self._listen_topic('__all__', callback)

    def broadcast(self, topics: str or List[str], *args, **kwargs):
        """
        Launch broadcast on the specify topic
        """
        if type(topics) == str:
            self._broadcast_topic(topics, *args, **kwargs)
        elif type(topics) == list:
            for topic in topics:
                self._broadcast_topic(topic, *args, **kwargs)
        else:
            raise ValueError("Unknown broadcast-service error, please submit "
                             "issue to https://github.com/Undertone0809/broadcast-service/issues")

    def broadcast_all(self, *args, **kwargs):
        """
        All topics listened on will be called back.
        Attention: Not all callback function will called. If your publish
         and your subscribe takes different arguments, your callback function
         will not be executed.
        """
        for topic in self.pubsub_channels.keys():
            self._broadcast_topic(topic, *args, **kwargs)

    def _listen_topic(self, topic_name: str, callback: Callable):
        if topic_name not in self.pubsub_channels.keys():
            self.pubsub_channels[topic_name] = []

        if callback not in self.pubsub_channels[topic_name]:
            self.pubsub_channels[topic_name].append(callback)

    def _broadcast_topic(self, topic_name: str, *args, **kwargs):
        """
        broadcast single topic.
        TODO fix problem: There is no guarantee that every callback function will be executed unnecessarily in some cases.
        """
        if topic_name not in self.pubsub_channels.keys():
            self.pubsub_channels[topic_name] = []

        for item in self.pubsub_channels[topic_name]:
            if self.enable_async:
                self.thread_pool.submit(
                    item, *args, **kwargs)
            else:
                item(*args, **kwargs)

        for item in self.pubsub_channels['__all__']:
            if self.enable_async:
                self.thread_pool.submit(
                    item, *args, **kwargs)
            else:
                item(*args, **kwargs)

    def stop_listen(self, topic_name: str, callback: Callable):
        if topic_name not in self.pubsub_channels.keys():
            raise RuntimeError("you didn't listen the topic:", topic_name)
        if callback not in self.pubsub_channels[topic_name]:
            pass
        else:
            self.pubsub_channels[topic_name].remove(callback)

    def on_listen(self, topics: str or Optional[List[str]] = None) -> Callable:
        """
        Decorator to listen specify topic. If topics is none, then listen all topics has exits.
        :param topics: topic list, you can input topic like: ["topic1", "topic2"].

        Usage::
            @broadcast_service.on_listen('topic1')
            def handle_all_msg():
                # your code

            @broadcast_service.on_listen(['topic1'])
            def handle_all_msg():
                # your code

            @broadcast_service.on_listen(['topic1','topic2'])
            def handle_all_msg():
                # your code

            @broadcast_service.on_listen()
            def handle_all_msg(*args, **kwargs):
                # your code

        Attention: Your params should keep '*args, **kwargs'. If you publish a topic take arguments,
        the callback function you handle should take arguments, otherwise it will not be called back.
        """
        def decorator(fn: Callable) -> Callable:
            if not topics:
                self._listen_topic('__all__', fn)
            elif type(topics) == str:
                self._listen_topic(topics, fn)
            elif type(topics) == list:
                self.listen(topics, fn)

            def inner(*args, **kwargs) -> Callable:
                ret = fn(*args, **kwargs)
                return ret
            return inner
        return decorator


broadcast_service = BroadcastService()



# Copyright 2022 Zeeland(https://github.com/Undertone0809/). All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from unittest import TestCase
from broadcast_service import broadcast_service


def wait(seconds=0.1):
    time.sleep(seconds)


class TestBroadcast(TestCase):
    def test_listen_of_common(self):
        self.test_listen_of_common_no_params = False
        self.test_listen_of_common_specify_params = False
        self.test_listen_multi_topic1_of_common = False
        self.test_listen_multi_topic2_of_common = False
        self.counter = 0
        self.all_counter = 0

        def handle_topic_no_params():
            self.test_listen_of_common_no_params = True

        def handle_topic_specify_params(a, b, c):
            self.assertEqual(11, a)
            self.assertEqual(22, b)
            self.assertEqual(33, c)
            self.test_listen_of_common_specify_params = True

        def handle_multi_topics(*args, **kwargs):
            self.counter += 1
            if args[0] == 111:
                self.test_listen_multi_topic1_of_common = True
            if args[0] == 222:
                self.test_listen_multi_topic2_of_common = True

        def handle_all_topics(*args, **kwargs):
            self.all_counter += 1

        # test all topics
        broadcast_service.subscribe_all(handle_all_topics)

        # test listen of common no params
        broadcast_service.subscribe(
            "test_listen_of_common_no_params", handle_topic_no_params)
        broadcast_service.publish("test_listen_of_common_no_params")
        wait()
        self.assertTrue(self.test_listen_of_common_no_params)

        # test listen of common specify params
        broadcast_service.subscribe(
            "test_listen_of_common_specify_params", handle_topic_specify_params)
        broadcast_service.publish(
            "test_listen_of_common_specify_params", 11, 22, 33)
        wait()
        self.assertTrue(self.test_listen_of_common_specify_params)

        # test listen multi topics of common
        topics = ['test_listen_multi_topic1_of_common', 'test_listen_multi_topic2_of_common']
        broadcast_service.subscribe(topics, handle_multi_topics)
        broadcast_service.publish("test_listen_multi_topic1_of_common", 111)
        broadcast_service.publish("test_listen_multi_topic2_of_common", 222)
        wait()
        self.assertTrue(self.test_listen_multi_topic1_of_common)
        self.assertTrue(self.test_listen_multi_topic2_of_common)
        self.assertEqual(2, self.counter)

        self.assertEqual(4, self.all_counter)

    def test_listen_of_decorator(self):
        self.test_listen_of_decorator_no_params1 = False
        self.test_listen_of_decorator_no_params2 = False
        self.test_listen_of_decorator_multi_params = False
        self.test_listen_of_decorator_specify_params = False
        self.test_listen_of_decorator_listen_all = False
        self.counter = 0

        @broadcast_service.on_listen("test_listen_of_decorator_no_params1")
        def handle_topic_no_params():
            self.test_listen_of_decorator_no_params1 = True
        
        @broadcast_service.on_listen(["test_listen_of_decorator_no_params2"])
        def handle_topic_no_params():
            self.test_listen_of_decorator_no_params2 = True

        @broadcast_service.on_listen(["test_listen_of_decorator_no_params1","test_listen_of_decorator_no_params2"])
        def handle_topic_no_params():
            if self.test_listen_of_decorator_no_params1 and self.test_listen_of_decorator_no_params2:
                self.test_listen_of_decorator_multi_params = True

        @broadcast_service.on_listen(["test_listen_of_decorator_specify_params"])
        def handle_topic_specify_params(a, b, c):
            self.assertEqual(11, a)
            self.assertEqual(22, b)
            self.assertEqual(33, c)
            self.test_listen_of_decorator_specify_params = True

        @broadcast_service.on_listen()
        def handle_listen_all_topics(*args, **kwargs):
            self.counter += 1
            self.test_listen_of_decorator_listen_all = True

        broadcast_service.publish("test_listen_of_decorator_no_params1")
        wait()
        self.assertTrue(self.test_listen_of_decorator_no_params1)

        broadcast_service.publish("test_listen_of_decorator_no_params2")
        wait()
        self.assertTrue(self.test_listen_of_decorator_no_params2)

        wait()
        self.assertTrue(self.test_listen_of_decorator_multi_params)

        broadcast_service.publish(
            "test_listen_of_decorator_specify_params", 11, 22, 33)
        wait()
        self.assertTrue(self.test_listen_of_decorator_specify_params)

        broadcast_service.publish("test_listen_of_decorator_listen_all")
        wait()
        self.assertTrue(self.test_listen_of_decorator_listen_all)
        self.assertEqual(4, self.counter)

    def test_listen_of_lambda(self):
        self.test_listen_of_lambda_no_params = False
        self.test_listen_of_lambda_specify_params = False

        def handle_topic_no_params():
            self.test_listen_of_lambda_no_params = True

        def handle_topic_specify_params(params: bool):
            self.test_listen_of_lambda_specify_params = params

        broadcast_service.subscribe(
            "test_listen_of_lambda_no_params", lambda: handle_topic_no_params())
        broadcast_service.publish("test_listen_of_lambda_no_params")
        wait()
        self.assertTrue(self.test_listen_of_lambda_no_params)

        broadcast_service.subscribe(
            "test_listen_of_lambda_specify_params", lambda: handle_topic_specify_params(True))
        broadcast_service.publish("test_listen_of_lambda_specify_params")
        wait()
        self.assertTrue(self.test_listen_of_lambda_no_params)

    def test_broadcast(self):
        self.test_broadcast_one_topic1 = False
        self.test_broadcast_one_topic2 = False
        self.test_broadcast_multi_topic1 = False
        self.test_broadcast_multi_topic2 = False
        self.counter = 0

        @broadcast_service.on_listen("test_broadcast_one_topic1")
        def handle_one_topic1():
            self.counter += 1
            self.test_broadcast_one_topic1 = True

        @broadcast_service.on_listen("test_broadcast_one_topic2")
        def handle_one_topic2():
            self.counter += 1
            self.test_broadcast_one_topic2 = True

        @broadcast_service.on_listen("test_broadcast_multi_topic1")
        def handle_multi_topic1():
            self.counter += 1
            self.test_broadcast_multi_topic1 = True

        @broadcast_service.on_listen("test_broadcast_multi_topic2")
        def handle_multi_topic2():
            self.counter += 1
            self.test_broadcast_multi_topic2 = True

        broadcast_service.publish("test_broadcast_one_topic1")
        wait()
        self.assertTrue(self.test_broadcast_one_topic1)

        broadcast_service.publish(["test_broadcast_one_topic2"])
        wait()
        self.assertTrue(self.test_broadcast_one_topic2)

        broadcast_service.publish(["test_broadcast_multi_topic1", "test_broadcast_multi_topic2"])
        wait()
        self.assertTrue(self.test_broadcast_multi_topic1 and self.test_broadcast_multi_topic2)

        broadcast_service.publish_all()
        wait()
        self.assertEqual(8, self.counter)



# Copyright 2022 Zeeland(https://github.com/Undertone0809/). All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import unittest
from broadcast_service import broadcast_service


def handle():
    time.sleep(1)


class TestAsync(unittest.TestCase):
    def test_async(self):
        start_time = time.time()
        broadcast_service.enable_async = True
        broadcast_service.listen("test_topic", handle)
        broadcast_service.broadcast("test_topic")
        used_time = time.time() - start_time
        self.assertLessEqual(used_time, 1)

    def test_sync(self):
        start_time = time.time()
        broadcast_service.enable_async = False
        broadcast_service.listen("test_topic", handle)
        broadcast_service.broadcast("test_topic")
        used_time = time.time() - start_time
        self.assertEqual(int(used_time), 1)


if __name__ == '__main__':
    unittest.main()
